import { PabPuissance, PabPuissanceParams } from "../pab/pab_puissance";
import { ParamCalculability, ParamDefinition } from "../param/param-definition";
import { Result } from "../util/result";
import { CloisonsParams } from "./cloisons_params";
import { ParallelStructure } from "./parallel_structure";
import { StructureKiviParams } from "./structure_kivi";

export { CloisonsParams };

export class Cloisons extends ParallelStructure {
    constructor(prms: CloisonsParams, dbg: boolean = false) {
        super(prms, dbg);
    }

    /**
     * paramètres castés au bon type
     */
    get prms(): CloisonsParams {
        return this._prms as CloisonsParams;
    }

    /**
     * Calcul du débit total, de la cote amont ou aval ou d'un paramètre d'une structure
     * @param sVarCalc Nom du paramètre à calculer :
     *                 "Q", "Z1", "DH" ou "n.X" avec "n" l'index de l'ouvrage et "X" son paramètre
     * @param rInit Valeur initiale
     * @param rPrec Précision attendue
     */
    public Calc(sVarCalc: string, rInit?: number, rPrec: number = 0.001): Result {
        // Mise à jour de ZRAM pour Kivi à partir de Z1 - PB
        this.updateKiviZRAM();

        // Transformation DH => Z2
        this.prms.Z2.v = this.prms.Z1.v - this.prms.DH.v;
        let sVC: string = sVarCalc;
        if (sVarCalc === "DH") {
            sVC = "Z2";
        }

        const r: Result = super.Calc(sVC, rInit, rPrec);

        // Transformation Z2 => DH
        if (sVarCalc === "DH") {
            r.vCalc = this.prms.Z1.v - r.vCalc;
        }

        // Ajout du calcul de la puissance dissipée
        const prms = this.getcalculatedparams(sVarCalc, r);
        const puissanceNub: PabPuissance = new PabPuissance(
            new PabPuissanceParams(
                prms.DH,
                prms.Q,
                prms.LB * prms.BB * prms.PB
            )
        );
        r.extraResults.PV = puissanceNub.Calc("PV", 0).vCalc;

        return r;
    }

    /**
     * paramétrage de la calculabilité des paramètres
     */
    protected setParametersCalculability() {
        super.setParametersCalculability();
        this.prms.LB.calculability = ParamCalculability.FREE;
        this.prms.BB.calculability = ParamCalculability.FREE;
        this.prms.PB.calculability = ParamCalculability.FREE;
        this.prms.DH.calculability = ParamCalculability.DICHO;
    }

    protected getcalculatedparams(sVarCalc: string, result: Result): { [key: string]: number } {
        const cPrms: { [key: string]: number } = {};
        for (const p of this.parameterIterator) {
            if (p.symbol === sVarCalc) {
                cPrms[p.symbol] = result.vCalc;
            } else {
                cPrms[p.symbol] = p.v;
            }
        }
        return cPrms;
    }

    private updateKiviZRAM() {
        for (const structure of this._structures) {
            if (structure.prms instanceof StructureKiviParams) {
                structure.prms.ZRAM.v = this.prms.Z1.v - this.prms.PB.v;
            }
        }
    }
}
